/**************************************************************************/
/*  ProuveParser: parser frontend for the PROUVÉ language                */
/*                                                                        */
/*  Copyright (C) 2004-2009 Ralf Treinen                                  */
/*                                                                        */
/*  This library is free software; you can redistribute it and/or         */
/*  modify it under the terms of the GNU Lesser General Public            */
/*  License as published by the Free Software Foundation; either          */
/*  version 2.1 of the License, or (at your option) any later version.    */
/*                                                                        */
/*  This library is distributed in the hope that it will be useful,       */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     */
/*  Lesser General Public License for more details.                       */
/*                                                                        */
/*  You should have received a copy of the GNU Lesser General Public      */
/*  License along with this library; if not, write to the Free Software   */
/*  Foundation, Inc., 59 Temple Place, Suite 330,                         */
/*  Boston, MA  02111-1307  USA                                           */
/*                                                                        */
/*  Ralf Treinen <treinen@pps.jussieu.fr>                                 */
/*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   */
/*  Université Paris Diderot, Paris, France                              */
/*  http://www.pps.jussieu.fr/~treinen                                    */
/*                                                                        */
/*  This software was mainly written while the author was with the        */
/*  Laboratoire Spécification et Vérification (LSV),                    */
/*  ENS de Cachan, Cachan, France                                         */
/**************************************************************************/

%{
  open Types
  open Signatures
  open Expressions
  open Declarations
  open Roles
  open Intervals
  open Scenarios
  open Specs
  open Formulas
%}

%token <string> IDENT TYPEVAR
%token <int> INTEGER
%token <string> BINOPA BINOPB BINOPC BINOPD BINOPE BINOPF
%token <string> STRING

%token ROLE BEGIN END DECLARE MUTABLE  
%token SEND RECV NEW MATCH WITH INV FAIL
%token ASSIGN SC COLON COMMA DOT PARL PARR BRAL BRAR ACCL ACCR DBRAL DBRAR
%token THEEND VBAR ARROW
%token IF THEN ELSE FI CASE OF ESAC CHOICE
%token SIGNATURE AXIOMS VARDECL HASHFUNCTIONS
%token SCENARIO FOREVER SPAWN PARALLEL FORALL EXISTS STAR
%token USES UNIQUE PUBLIC INITIAL ALWAYS AND NEGATION ISSECRET AT BOX

%start spec
%type <Specs.spec> spec
%start assertion
%type <Assertions.assertion> assertion

/* Infix operators in increasing order of precedence */
%left BINOPF
%left BINOPE
%left BINOPD
%left BINOPC
%left BINOPB
%left BINOPA
%left DOT
%right ARROW

%%

spec:
  optional_signature optional_axioms rolelist optional_declaration 
  scenario THEEND 
  { ($1,$2,$3,$4,$5) }
;

/**********************************************************************/
/*                          Signatures                                */
/**********************************************************************/

optional_signature:
  SIGNATURE siglist END                         { ($2,[]) }
| SIGNATURE siglist HASHFUNCTIONS siglist END   { ($2,$4) }
| /* empty */                                   { ([],[]) }
;

siglist:
  fundecl SC siglist  { $1@$3 }
| /* empty */         { [] }
;

fundecl:
  non_empty_identlist COLON typeexp
    { (flatten_signature $1 $3) }
;

/*********************************************************************/
/*                     Type expressions                              */
/*********************************************************************/

/* type expressions */
typeexp:
  IDENT                                   { Typeconst($1)}
| TYPEVAR                                 { Typevar($1) } 
| IDENT PARL non_empty_typeexp_list PARR  { Typeapply($1,$3) }
| typeexp ARROW typeexp                   { Typearrow($1,$3) }
| PARL non_empty_typeexp_list PARR        { Typetuple($2) }
;

/* non-empty lists of type expressions */
non_empty_typeexp_list:
  typeexp                                   { [$1] }
| typeexp COMMA non_empty_typeexp_list      { $1::$3 }
;


/*********************************************************************/
/*                         Axioms                                    */
/*********************************************************************/

optional_axioms:
  AXIOMS axiomlist END                                     { [], $2 }
| AXIOMS DECLARE terminated_decllist BEGIN axiomlist END   { $3, $5 }
| /* empty */                                              { [], [] }
;

axiomlist:
  expression SC axiomlist      { if not (Axioms.is_axiom $1)
				 then raise Parsing.Parse_error;
                                 $1::$3 }
| /* empty */                  { [] }
;

/*********************************************************************/
/*                          Roles                                    */
/*********************************************************************/

/* list of roles */
rolelist:
  role rolelist   {$1::$2}
| /* empty */     {[]}
;

/* one role */
role:
  ROLE IDENT PARL unterminated_decllist PARR block  
    { Role($2,$4,process_instblock $2 $6) }
| ROLE IDENT PARL PARR block                        
    { Role($2,[],process_instblock $2 $5) }
;

/* either the non-empty list of declarations without a terminating ";" */
/* or the empty list of declarations.                                  */
unterminated_decllist:
  declaration                             { $1 }
| declaration SC unterminated_decllist    { $1@$3 }
;

/* a possibly empty list of declarations that are all terminated */
terminated_decllist:
  declaration SC terminated_decllist    { $1@$3 }
| /* empty */                           { [] }
;


/* one declaration, several identifiers */
declaration:
  non_empty_identlist optional_typeexp    { (flatten_declaration $1 $2) }
;

/* an optional type expression */
optional_typeexp:
  COLON typeexp             { (Logical,$2) }
| COLON MUTABLE typeexp     { (Mutable,$3) }
| /* empty */               { (Logical,Typeconst("message")) }
| COLON MUTABLE             { (Mutable,Typeconst("message")) }
; 

/* non-empty list of identifiers */
non_empty_identlist:
  IDENT                             { [$1] }
| IDENT COMMA non_empty_identlist   { $1::$3 }
;

/* block of instructions */
block:
  DECLARE terminated_decllist BEGIN instlist END       { ($2,$4) }
| BEGIN instlist END                                   { ([],$2) }
;

/*list of instructions */
instlist:
  instruction SC instlist  { $1::$3 }
| /* empty */           { [] }
;

/* instruction possibly with a label */
instruction:
  IDENT COLON naked_instruction                  { IRawLabeled($1,$3) }
| naked_instruction                              { $1 }
;

/* naked instruction: without label */
naked_instruction:
  block                                          { IRawBlock($1) }
| SEND PARL expression PARR                      { IRawSend($3) }
| RECV PARL patterns PARR                        { IRawRecv($3) }
| FAIL                                           { IRawFail }
| IDENT ASSIGN expression                        { IRawAssign($1,$3) }
| NEW PARL IDENT PARR                            { IRawNew($3) }
| IDENT PARL expression COMMA IDENT PARR         { if ($1="makekey")
						   then IRawMakeKey($3,$5)
						   else
						     raise Parsing.Parse_error}
| MATCH expression WITH patterns END             { IRawMatch($2,$4) }
| CHOICE choicelist END                          { IRawChoice($2) }
| IF expression THEN instlist ELSE instlist FI   { IRawConditional($2,$4,$6) }
| IF expression THEN instlist FI                 { IRawConditional($2,$4,[]) }
| CASE expression OF caselist ELSE instlist ESAC { IRawCase($2,$4,$6) }

/* list of cases */
caselist:
  case SC              { [$1] }
| case SC caselist     { $1::$3 }
;

/* one case */
case:
  patterns ARROW instlist { ($1,$3) }
;

/* list of choices */
choicelist:
   instlist                   { [$1] }
|  instlist VBAR choicelist   { $1::$3 }
;

/**********************************************************************/
/*                        Expressions                                 */
/**********************************************************************/
expression:
  INTEGER                        { Int(undef_info_exp,$1) }
| IDENT                          { Ident(undef_info_exp,$1) }
| expression BINOPA expression   { Infix(undef_info_exp,$2,$1,$3) }
| expression BINOPB expression   { Infix(undef_info_exp,$2,$1,$3) }
| expression BINOPC expression   { Infix(undef_info_exp,$2,$1,$3) }
| expression BINOPD expression   { Infix(undef_info_exp,$2,$1,$3) }
| expression BINOPE expression   { Infix(undef_info_exp,$2,$1,$3) }
| expression BINOPF expression   { Infix(undef_info_exp,$2,$1,$3) }
| expression DOT expression      { Infix(undef_info_exp,".",$1,$3) }
| IDENT PARL exprlist PARR       { Apply(undef_info_exp,$1,$3) }
| INV PARL expression PARR       { Inverse(undef_info_exp,$3) }
| BRAL expression COMMA exprlist BRAR  { Tuple(undef_info_exp,$2::$4) }
| DBRAL exprlist DBRAR           { Liste(undef_info_exp,$2) }
| DBRAL DBRAR                    { Liste(undef_info_exp,[]) }
| PARL expression PARR           { $2 }
| ACCL assoc_list ACCR           { AssocList(undef_info_exp,$2) }
;

/* non-empty list of expressions */
exprlist:
  expression                   { [$1] }
| expression COMMA exprlist    { $1::$3 }
;

/* association list */
assoc_list:
  /* empty */ { [] }
| ne_assoc_list { $1}
;

/* non-empty association list */
ne_assoc_list:
  expression COLON expression                      { [($1,$3)] }
| expression COLON expression COMMA ne_assoc_list  { ($1,$3)::$5 }
;
  
/**********************************************************************/
/*                      Patterns                                      */
/**********************************************************************/

/* non-empty list of patterns */
patterns:
  pattern                  { [$1] }
| pattern SC patterns    { $1::$3 } 
;

/* one single pattern */
pattern:
  expression                          { $1 }
| PARL expression COLON typeexp PARR  { $2 }
;

/***********************************************************************/
/*                 Global Variable Declaration                         */
/***********************************************************************/
optional_declaration:
  /* empty */                              { [] }
| VARDECL skolem_decllist END              { $2 }
;

skolem_decllist:
  /* empty */                              { [] }
| skolem_declaration SC skolem_decllist    { $1@$3 }
;

skolem_declaration:
  skolem_identlist optional_typeexp    { (flatten_sk_declaration $1 $2) }
;

/* non-empty list of identifiers */
skolem_identlist:
  IDENT                                                      { [$1,[]] }
| IDENT PARL non_empty_typeexp_list PARR                     { [$1,$3] }
| IDENT PARL non_empty_typeexp_list PARR COMMA skolem_identlist
      { ($1,$3)::$6 }
| IDENT COMMA skolem_identlist                               { ($1,[])::$3 }
;

/**********************************************************************/
/*                       Scenarios                                    */
/**********************************************************************/

scenario:
  SCENARIO sceneinst END         { process_scenario "Scene" $2 }
;

/* scenario instruction with an optional label */
sceneinst:
  IDENT COLON naked_sceneinst                  { SRawLabeled($1,$3) }
| naked_sceneinst                              { $1 }
;

/* scenario instruction without label */
naked_sceneinst:
  IDENT PARL exprlist PARR                        { SRawCall($1,$3) }
| IDENT PARL PARR                                 { SRawCall($1,[]) }
| NEW PARL expression PARR                        { SRawNew($3) }
| expression ASSIGN expression                    { SRawAssign($1,$3) }
| IF expression THEN scenelist ELSE scenelist FI  { SRawConditional($2,$4,$6) }
| IF expression THEN scenelist FI                 { SRawConditional($2,$4,[]) }
| FOREVER scenelist END                           { SRawForever($2) }
| PARALLEL sceneparlist END                       { SRawParallel($2) }
| BEGIN scenelist END                             { SRawSequence($2) }
| FORALL valuedec DOT sceneinst                   { SRawForall($2,[],$4) }
| FORALL valuedec BRAL interval_list BRAR DOT sceneinst
                                                  { SRawForall($2,$4,$7) }
| EXISTS valuedec DOT sceneinst                   { SRawExists($2,$4) }
| DECLARE terminated_decllist BEGIN scenelist END { SRawBlock($2,$4) } 
;

valuedec:
  non_empty_identlist COLON typeexp 
      { make_dec_logical (flatten_signature $1 $3) }
| non_empty_identlist COLON typeexp SC valuedec
      { (make_dec_logical (flatten_signature $1 $3)) @ $5 }
;


/* sequential composition in a scenario */
scenelist:
  sceneinst SC scenelist   { $1::$3 }
| sceneinst                { [$1] }
| sceneinst SC             { [$1] }
;

/* parallel composition in a scenario */
sceneparlist:
  sceneinst VBAR sceneparlist  { $1::$3 } 
| sceneinst                    { [$1] }
;

/*********************************************************************/
/*                  Interval Lists                                   */
/*********************************************************************/

interval_list: 
  interval                                        { [ $1 ] }
| interval COMMA interval_list                    { $1 :: $3 }
;

interval:
  omint                                           { ($1,$1) }
| omint DOT DOT omint                             { ($1,$4) }
;

omint:
  INTEGER                                         { OmInt($1) }
| STAR                                            { Omega }
;


/*********************************************************************/
/*                  Assertions                                       */
/*********************************************************************/

assertion:
  USES STRING optional_public optional_initial ALWAYS formula THEEND
                                                  { $2,$3,$4,$6 }
;

optional_public:
  /* empty */                                     { [] }
| PUBLIC                                          { [] }
| PUBLIC non_empty_identlist                      { $2 }
;

optional_initial:
  /* empty */                                     { Ftrue }
| INITIAL formula                                 { $2 }
;

/*********************************************************************/
/*                  Formulas                                         */
/*********************************************************************/

formula:
  FORALL valuedec DOT formula                     { Fforall($2,$4) }
| EXISTS valuedec DOT formula                     { Fexists($2,$4) }
| UNIQUE valuedec DOT formula                     { Funique($2,$4) }
| BOX formula                                     { Fbox($2) }
| formulb                                         { $1 }   
; 

formulb:
  formulb VBAR formulc                            { For($1,$3) } 
| formulb ARROW formulc                           { Fimplies($1,$3) }
| formulc                                         { $1 }
;

formulc:
  formulc AND formuld                             { Fand($1,$3) }
| formuld                                         { $1 }
  
formuld:
  expression                                      { Fatom($1) } 
| PARL formula PARR                               { $2 }
| ISSECRET PARL expression PARR                   { Fsecret($3) }
| IDENT AT IDENT                                  { Fprocessat($1,$3) }
| AT IDENT                                        { Fscenarioat $2 }
;




